__all__ = ['optimizealpha']
import sys
import logging
import shapely
from shapely.geometry import MultiPoint
import numpy
import alphashape


def _testalpha(alpha, points):
    """
    Test
    """
    polygon = alphashape.alphashape(points, alpha)
    if isinstance(polygon, shapely.geometry.polygon.Polygon) and all(
            [polygon.intersects(point) for point in points]):
        return polygon.area
    else:
        return float('inf')


def optimizealpha(points, max_iterations=10000):
    """
    Solve for the alpha parameter.

    Attempt to determine the alpha parameter that best wraps the given set of
    points in one polygon without dropping any points.

    Note:  If the solver fails to find a solution, a value of zero will be
    returned, which when used with the alphashape function will safely return a
    convex hull around the points.

    Args:

        points (list): an iterable container of points
        max_iterations (int): maximum number of iterations while finding the
            solution

    Returns:

        float: The optimized alpha parameter

    """
    if not isinstance(points, MultiPoint):
        points = MultiPoint(list(points))
    # Set the bounds
    lower = 0.

    # Ensure the upper limit bounds the solution
    upper = sys.float_info.max
    if _testalpha(upper, points) != float('inf'):
        logging.error('the max float value does not bound the alpha '
                      'parameter solution')
        return 0.
    counter = 0
    while (upper - lower) > numpy.finfo(float).eps * 2:
        test_alpha = (upper + lower) * .5
        if _testalpha(test_alpha, points) == float('inf'):
            upper = test_alpha
        else:
            lower = test_alpha
        counter += 1
        if counter > max_iterations:
            logging.warning('maximum allowed iterations reached while '
                            'optimizing the alpha parameter')
            break
    return lower



"""
Tools for working with alpha shapes.
"""
__all__ = ['alphashape']

import itertools
import math
from shapely.ops import cascaded_union, polygonize
from shapely.geometry import MultiPoint, MultiLineString
from scipy.spatial import Delaunay
import numpy as np
import alphashape


def alphashape(points, alpha=None):
    """
    Compute the alpha shape (concave hull) of a set of points.  If the number
    of points in the input is three or less, the convex hull is returned to the
    user.  For two points, the convex hull collapses to a `LineString`; for one
    point, a `Point`.

    Args:

        points (list): an iterable container of points

        alpha (float): alpha value

    Returns:

        ``shapely.geometry.Polygon`` or ``shapely.geometry.LineString`` or
        ``shapely.geometry.Point``: the resulting geometry
    """
    if not isinstance(points, MultiPoint):
        points = MultiPoint(list(points))

    # If given a triangle for input, or an alpha value of zero or less,
    # return the convex hull.
    if len(points) < 4 or (alpha is not None and alpha <= 0):
        return points.convex_hull

    # Determine alpha parameter if one is not given
    if alpha is None:
        alpha = alphashape.optimizealpha(points)

    coords = np.array([point.coords[0] for point in points])
    tri = Delaunay(coords)
    edges = set()
    edge_points = []

    # Loop over triangles
    for ia, ib, ic in tri.vertices:
        pa = coords[ia]
        pb = coords[ib]
        pc = coords[ic]

        # Lengths of sides of triangle
        a = math.sqrt((pa[0] - pb[0])**2 + (pa[1] - pb[1])**2)
        b = math.sqrt((pb[0] - pc[0])**2 + (pb[1] - pc[1])**2)
        c = math.sqrt((pc[0] - pa[0])**2 + (pc[1] - pa[1])**2)

        # Semiperimeter of triangle
        s = (a + b + c) * 0.5

        # Area of triangle by Heron's formula
        area = math.sqrt(s * (s - a) * (s - b) * (s - c))

        # Radius Filter
        print(alpha)
        if area > 0 and a * b * c / (4.0 * area) < 1.0 / alpha:
            for i, j in itertools.combinations([ia, ib, ic], r=2):
                if (i, j) not in edges and (j, i) not in edges:
                    edges.add((i, j))
                    edge_points.append(coords[[i, j]])

    m = MultiLineString(edge_points)
    triangles = list(polygonize(m))
    return cascaded_union(triangles)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `alphashape` package."""


import unittest
import logging
from click.testing import CliRunner

import shapely
from alphashape.alphashape import alphashape
from alphashape import cli


class TestAlphashape(unittest.TestCase):
    """Tests for `alphashape` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_given_a_point_return_a_point(self):
        """
        Given a point, the alphashape function should return the same point
        """
        assert shapely.geometry.Point([0., 0.]) == alphashape([(0., 0.)], 0)
        assert shapely.geometry.Point([1., 0.]) == alphashape([(1., 0.)], 0)
        assert shapely.geometry.Point([0., 1.]) == alphashape([(0., 1.)], 0)
        assert shapely.geometry.Point([0., 0.]) == alphashape([(0., 0.)], 99)
        assert shapely.geometry.Point([1., 0.]) == alphashape([(1., 0.)], 99)
        assert shapely.geometry.Point([0., 1.]) == alphashape([(0., 1.)], 99)

    def test_given_a_line_with_dupicate_points_return_a_point(self):
        """
        Given a line with duplicate points, the alphashape function should
        return a point
        """
        print(alphashape)
        logging.info(alphashape)
        alphashape([(0., 1.), (0., 1.)], 0)
        assert shapely.geometry.Point([0., 1.]) == alphashape(
            [(0., 1.), (0., 1.)], 0)

    def test_given_a_line_with_unique_points_return_a_line(self):
        """
        Given a line with unique points, the alphashape function should return
        the same line
        """
        assert shapely.geometry.LineString([(0., 0.), (0., 1.)]) == alphashape(
            [(0., 0.), (0., 1.)], 0)
        assert shapely.geometry.LineString([(1., 0.), (0., 1.)]) == alphashape(
            [(1., 0.), (0., 1.)], 0)

    def test_given_a_triangle_with_duplicate_points_returns_a_point(self):
        """
        Given a triangle with two unique points, the alphashape function should
        return a point
        """
        assert shapely.geometry.Point((0., 1.)) == alphashape(
            [(0., 1.), (0., 1.), (0., 1.)], 0)

    def test_given_a_triangle_with_two_duplicate_points_returns_a_line(self):
        """
        Given a line with two unique points, the alphashape function should
        return a line with the unique points
        """
        assert shapely.geometry.LineString([(1., 0.), (0., 1.)]) == alphashape(
            [(1., 0.), (0., 1.), (0., 1.)], 0)

    def test_given_a_four_point_polygon_with_small_alpha_return_input(self):
        """
        Given a polygon with four points, and an alpha value of zero, return
        the input as a polygon.
        """
        assert shapely.geometry.Polygon([
            (0., 0.), (0., 1.), (1., 1.), (1., 0.), (0., 0.)]).equals(
                alphashape([(0., 0.), (0., 1.), (1., 1.), (1., 0.)], 1.e-9))

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        assert 'alphashape.cli.main' in result.output
        help_result = runner.invoke(cli.main, ['--help'])
        assert help_result.exit_code == 0
        assert '--help  Show this message and exit.' in help_result.output



"""Top-level package for Alpha Shape Toolbox."""

__author__ = """Kenneth E. Bellock"""
__email__ = 'ken@bellock.net'

from .alphashape import alphashape
from .optimizealpha import optimizealpha
from ._version import __version__  # noqa: F401
__all__ = ['alphashape', 'optimizealpha']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `alphashape` package."""


import unittest
import logging

import shapely
from alphashape import optimizealpha


class TestOptimizeAlapha(unittest.TestCase):
    """Tests for `alphashape` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_given_a_point_return_a_point(self):
        """
        Given a point, the alphashape function should return the same point
        """
        alpha = optimizealpha(
            [(0., 0.), (0., 1.), (1., 1.), (1., 0.),
             (0.5, 0.25), (0.5, 0.75), (0.25, 0.5), (0.75, 0.5)])
        assert alpha > 3. and alpha < 3.5
